#### Return the NML for a matrix-BN pair

## Author: Phillip P.A. Staniczenko  pstaniczenko@sesync.org

#### Required files
## fNMLGet executable

#### Inputs
## "Zipped" community distribution matrix
## Bayesian network

#### Outputs
## Total length for model-data combination

# Clear workspace
rm(list=ls(all=TRUE))

#### Libraries
require(gsl)
library(igraph)

#### Variables

BN_model <- 3  # 0: AND
			   # 1: OR
			   # 2: SUM
			   # 3: FULL
					 
#### Data

mybipartitematrixzipped <- "mybipartitematrix.txt-z-41514-54"

myBN <- as.matrix(read.table("blakeDAG_CD_BN_mod_FULL.csv", sep=","))
myBN <- as.numeric(myBN[2:55, 2:55])
myBN[is.na(myBN)] <- 0
myBN <- matrix(myBN, 54, 54)
myBN <- abs(myBN)

modelname_BN <- "BN_Test"

# DAG loaded from file
myDAGcols <- myBN

myDAGcols_nodeorder <- 1:dim(myDAGcols)[2]

myDAG <- myDAGcols
myDAGorder <- myDAGcols_nodeorder

# REORDER
M <- abs(myDAG)
graph <- graph.adjacency(M, mode="directed",)
REORDER <- topo_sort(graph, mode="out")
REORDER <- as.numeric(REORDER)
#
mybipartitematrixzipped_temp <- read.table(mybipartitematrixzipped, sep="")
REORDER_zipped <- c(REORDER, max(REORDER) + 1)
mybipartitematrixzipped_temp <- mybipartitematrixzipped_temp[, REORDER_zipped]
mybipartitematrixzipped_REORDER <- paste(mybipartitematrixzipped, "_REORDER", sep="")
write.table(mybipartitematrixzipped_temp, mybipartitematrixzipped_REORDER, sep=" ", row.names=FALSE, col.names=FALSE)
#
myDAG <- myDAG[REORDER, REORDER]

#### Functions

# Compute complexity of random graph of size X
Complexity <- function(X){
  if (X == 0){
    return (0.0)
  }
  # for small X, use the exact value
  XX = X
  if (XX < 100.){
    return (log2(exp(XX + log(gamma_inc(XX,XX)) - (XX - 1.0) * log(XX)) + 1.0))
  }
  # for large X, use the approximation
  return (log2(1. + sqrt(pi * XX / 2.) - 1./3. + sqrt(2. * pi) / (24. * sqrt(XX)) 
               - 4. / (135. * XX)))
}

# Compute NML (really, total length) for a slice containing Ones ones and Zeros zeros
NMLSlice <- function(Ones, Zeros){
  NML = Complexity(Ones + Zeros)
  p = 0.0
  # compute maximum log likelihood
  p = Ones / (Ones + Zeros)
  NML = NML - Ones * log2(p) - Zeros * log2(1.0 - p)
  return (NML)
}

# Compute NML (really, total length) for a matrix-SBM combination
getNML <- function(mybipartitematrix, rowgroups, colgroups){
  # iterate over groups
  NMLmatrix <- 0.0
  for (i in 1:length(rowgroups)){
  	for (j in 1:length(colgroups)){
  	  Ones <- 0
  	  mycounter <- 0
  	  # iterate over elements within a rowgroup-colgroup pair
  	  for (k in 1:length(rowgroups[[i]])){
  	  	for (l in 1:length(colgroups[[j]])){
  	  	  mycounter <- mycounter + 1
  	  	  Ones <- Ones + mybipartitematrix[rowgroups[[i]][k],colgroups[[j]][l]]
  	  	}
  	  }
  	  NMLmatrix <- NMLmatrix + NMLSlice(Ones, mycounter - Ones)
  	}
  }
  return(NMLmatrix)
}

# Check if graph is DAG
isDAG <- function(myDAG){
  M <- myDAG
  M[is.na(M)] <- 0
  M[M == -1] = 1
  graph <- graph.adjacency(M, mode="directed",)
  return(is.dag(graph))
}

# Get DAG in format for fNMLGet executable
getreducedDAGform <- function(myDAG, myDAGorder){
  numnodes <- length(myDAGorder)
  reducedDAGform <- numnodes - myDAGorder
  errorcheck <- c(0,0)
  for (i in numnodes:2){
  	reducedDAGform <- c(reducedDAGform, rev(myDAG[1:(i - 1), i]))
  	errorcheck <- c(errorcheck, rev(myDAG[1:(i - 1), i]))
  }
  write.table(reducedDAGform, file="myDAGreduced.txt", row.names=FALSE, col.names=FALSE, sep=" ")
  return(errorcheck)
}

#### Main

print("RUN TOTAL LENGTH CALCULATION")
beginTime <- Sys.time()


if (TRUE){
  # Run code for BN
  DAGcheck <- isDAG(myDAGcols)
  if (DAGcheck==TRUE){
    print ("CHECK: DAG is acyclic")
    
    # write DAG in reduced form
    errorcheck <- getreducedDAGform(myDAG, myDAGorder)
    # call fNMLGet executable
    mypattern <- mybipartitematrixzipped_REORDER
    myfiles <- list.files(path=".", pattern=mypattern, full.names=FALSE)
    zipmatrixnumrows <- unlist(regmatches(myfiles[1], gregexpr('\\(?[0-9,]+', myfiles[1])))[1]
    zipmatrixnumcols <- unlist(regmatches(myfiles[1], gregexpr('\\(?[0-9,]+', myfiles[1])))[2]
    myexpression <- paste("./fNMLGet ", zipmatrixnumrows, " ", zipmatrixnumcols, " ", myfiles[1], " myDAGreduced.txt ", toString(BN_model), " 123 100 100", sep="")
    system(myexpression)
    myfiles2 <- list.files(path=".", pattern=mypattern, full.names=FALSE)
    BN_model_results <- unlist(regmatches(myfiles2[2], gregexpr('\\(?[0-9,.]+', myfiles2[2])))
    TotalLength <- as.numeric(BN_model_results[4])
    
    # Print output
    print("Model name:")
    print(modelname_BN)
    print("Logic type:")
    if (BN_model==0){
      print("AND")
    } else if (BN_model==1){
      print("OR")
    } else if (BN_model==2){
      print("SUM")
    } else if (BN_model==3){
      print("FULL")
    }
    print("Results for BN--")
    print("Total length:")
    print(TotalLength)
    
    # Clean up files
    system("rm myDAGreduced.txt")
    system(paste("rm ", mybipartitematrixzipped, "_REORDER", sep=""))
  } else {
    print ("ERROR: DAG is not acyclic, please modify")
  }
}

runTime <- Sys.time() - beginTime
print("Program completed successfully")
print("Run time:")
print(runTime)
if (sum(errorcheck) != sum(myDAG)){print("ERROR: Not all BN edges are being included when calculating Total Length!!!")}